@@ -58,7 +58,8 @@ async def init(data_dir, server_conf, send_queue, recv_queue):
         ssl=None if server_conf.ssl is True else False,
     )
 
-    app = web.Application()
+    # 100 MB max POST size
+    app = web.Application(client_max_size=1024 ** 2 * 100)
 
     app.add_routes(
         [
@@ -69,6 +70,10 @@ async def init(data_dir, server_conf, send_queue, recv_queue):
                 r"/_matrix/client/r0/rooms/{room_id}/send/{event_type}/{txnid}",
                 proxy.send_message,
             ),
+            web.post(
+                r"/_matrix/client/r0/rooms/{room_id}/send/{event_type}",
+                proxy.send_message,
+            ),
             web.post("/_matrix/client/r0/user/{user_id}/filter", proxy.filter),
             web.post("/.well-known/matrix/client", proxy.well_known),
             web.get("/.well-known/matrix/client", proxy.well_known),
@@ -88,6 +93,15 @@ async def init(data_dir, server_conf, send_queue, recv_queue):
                 "/_matrix/media/r0/download/{server_name}/{media_id}/{file_name}",
                 proxy.download,
             ),
+            web.post(
+                r"/_matrix/media/r0/upload",
+                proxy.upload,
+            ),
+            web.put(
+                r"/_matrix/client/r0/profile/{userId}/avatar_url",
+                proxy.profile,
+            ),
+
         ]
     )
     app.router.add_route("*", "/" + "{proxyPath:.*}", proxy.router)
@@ -245,7 +259,7 @@ async def daemon(context, log_level, debug_encryption, config, data_path):
         "connect to pantalaimon."
     )
 )
-@click.version_option(version="0.6.2", prog_name="pantalaimon")
+@click.version_option(version="0.9.1", prog_name="pantalaimon")
 @click.option(
     "--log-level",
     type=click.Choice(["error", "warning", "info", "debug"]),
